@@ -2,23 +2,19 @@
 PC-BASIC - interpreter.py
 BASIC interpreter
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
-import string
 import struct
 
 from .base import error
 from .base import tokens as tk
+from .base.tokens import DIGITS
 from .base import codestream
 from . import values
 
 
-# bytes constants
-DIGITS = string.digits
-
-
 class Interpreter(object):
     """BASIC interpreter."""
 
@@ -368,19 +364,19 @@ class Interpreter(object):
         """Initialise a FOR loop."""
         # read variable
         varname = self._memory.complete_name(next(args))
-        vartype = varname[-1]
-        start = values.to_type(vartype, next(args))
+        vartype = varname[-1:]
+        start = values.to_type(vartype, next(args)).clone()
         # only raised after the TO has been parsed
         if vartype in (values.STR, values.DBL):
             raise error.BASICError(error.TYPE_MISMATCH)
-        stop = values.to_type(vartype, next(args))
+        stop = values.to_type(vartype, next(args)).clone()
         step = next(args)
         if step is not None:
-            step = values.to_type(vartype, step)
+            step = values.to_type(vartype, step).clone()
         list(args)
         if step is None:
             # convert 1 to vartype
-            step = self._values.from_value(1, varname[-1])
+            step = self._values.from_value(1, varname[-1:])
         ins = self.get_codestream()
         # find NEXT
         forpos, nextpos = self._find_next(ins, varname)
@@ -389,7 +385,7 @@ class Interpreter(object):
         # obtain a view of the loop variable
         self.for_stack.append((varname, stop, step, step.sign(), forpos, nextpos,))
         # empty loop: jump to NEXT without executing block
-        if (start.gt(stop) if step.sign() > 0 else stop.gt(start)):
+        if (start.gt(stop) if step.sign() >= 0 else stop.gt(start)):
             ins.seek(nextpos)
             self.iterate_loop()
 
@@ -534,7 +530,7 @@ class Interpreter(object):
                 self._program_code.seek(current)
                 raise error.BASICError(error.OUT_OF_DATA)
             self._program_code.skip_blank()
-            if name[-1] == values.STR:
+            if name[-1:] == values.STR:
                 # for unquoted strings, payload starts at the first non-empty character
                 address = self._program_code.tell_address()
                 word = self._program_code.read_to((b',', b'"',) + tk.END_STATEMENT)
