@@ -2,7 +2,7 @@
 PC-BASIC - devicebase.py
 Devices, Files and I/O operations
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -12,17 +12,18 @@ import struct
 import logging
 from contextlib import contextmanager
 
+from ...compat import iterchar
 from ..base import error
 from ..base.eascii import as_bytes as ea
 from .. import values
 
 def nullstream():
-    return open(os.devnull, b'r+')
+    return open(os.devnull, 'r+b')
 
 
 # magic chars used by some devices to indicate file type
-TYPE_TO_MAGIC = { b'B': b'\xFF', b'P': b'\xFE', b'M': b'\xFD' }
-MAGIC_TO_TYPE = { b'\xFF': b'B', b'\xFE': b'P', b'\xFD': b'M' }
+TYPE_TO_MAGIC = {b'B': b'\xFF', b'P': b'\xFE', b'M': b'\xFD'}
+MAGIC_TO_TYPE = {b'\xFF': b'B', b'\xFE': b'P', b'\xFD': b'M'}
 
 
 
@@ -257,7 +258,7 @@ class TextFileBase(RawFile):
         to_read = num - len(self._readahead)
         if to_read > 0:
             with safe_io():
-                self._readahead.extend(list(self._fhandle.read(to_read)))
+                self._readahead.extend(iterchar(self._fhandle.read(to_read)))
         return b''.join(self._readahead[:num])
 
     def read(self, num):
@@ -272,7 +273,7 @@ class TextFileBase(RawFile):
         if len(output) <= 1:
             self._previous = self._current
         else:
-            self._previous = output[-2]
+            self._previous = output[-2:]
         self._current = output[-1:]
         return output
 
@@ -311,18 +312,20 @@ class TextFileBase(RawFile):
         s_width = 0
         newline = False
         # find width of first line in s
-        for c in s:
+        for c in iterchar(s):
             if c in (b'\r', b'\n'):
                 newline = True
                 break
             if c >= b' ':
                 # nonprinting characters including tabs are not counted for WIDTH
                 s_width += 1
-        if (can_break and self.width != 255 and self.col != 1 and
-                self.col-1 + s_width > self.width and not newline):
+        if (
+                can_break and self.width != 255 and self.col != 1 and
+                self.col-1 + s_width > self.width and not newline
+            ):
             self.write_line()
             self.col = 1
-        for c in s:
+        for c in iterchar(s):
             # don't replace CR or LF with CRLF when writing to files
             if c == b'\r':
                 self._fhandle.write(c)
@@ -336,7 +339,7 @@ class TextFileBase(RawFile):
                     if self.col == 257:
                         self.col = 1
 
-    def write_line(self, s=''):
+    def write_line(self, s=b''):
         """Write string and follow with device-standard line break."""
         self.write(s + b'\r')
 
@@ -425,7 +428,7 @@ class InputMixin(object):
         # skip trailing whitespace before any comma or hard separator
         if c and c in INPUT_WHITESPACE or (quoted and c == b'"'):
             self._skip_whitespace(b' ')
-            if (self.peek(1) in b',\r'):
+            if self.peek(1) in b',\r':
                 c = self.read_one()
         # file position is at one past the separator char
         return word, c
@@ -636,7 +639,7 @@ class SCRNFile(RawFile):
         s_width = 0
         newline = False
         # find width of first line in s
-        for c in s:
+        for c in iterchar(s):
             if c in (b'\r', b'\n'):
                 newline = True
                 break
@@ -652,7 +655,7 @@ class SCRNFile(RawFile):
             self.screen.write_line(do_echo=do_echo)
             self._col = 1
         cwidth = self.screen.mode.width
-        for c in s:
+        for c in iterchar(s):
             if self.width <= cwidth and self.col > self.width:
                 self.screen.write_line(do_echo=do_echo)
                 self._col = 1
